#!/usr/bin/python3.6
# -*- coding: utf-8 -*-

import torch
import argparse
from config import opt
from datasets import train_data_load, test_data_load, sample_test_data_load
from utils import confusion, demenstrate, logger, optim_factory, plot_feature, scheduler_factory, send_email, tools, train
import time
from torch import nn
import os
from torch.utils.tensorboard import SummaryWriter
from fine_tune import fine_tuning


from model.BSCP import BSCP_Resnet50
from model.BSCP_kernel import BSCP_kernelPooling


def parse_args():

    parser = argparse.ArgumentParser()

    parser.add_argument('--data_type', type=str, default=opt.data_type, help='Select dataset')
    parser.add_argument('--num_classes', type=int, default=opt.num_classes, help='The class to which this dataset corresponds')

    parser.add_argument('--model_type', type=str, default=opt.model_type, help='Select the model to use')
    parser.add_argument('--l1', type=int, default=512, help='U')
    parser.add_argument('--l2', type=int, default=512, help='V')
    parser.add_argument('--p', type=int, default=512, help='output dim')
    parser.add_argument('--rank', type=int, default=2, help='rank')

    parser.add_argument('--U', nargs='+', type=int, default=[512, 512, 512], help='U')
    parser.add_argument('--V', nargs='+', type=int, default=[512, 512, 512], help='V')
    parser.add_argument('--P', nargs='+', type=int, default=[512, 512, 512], help='P')
    parser.add_argument('--RANK', nargs='+', type=int, default=[2, 2, 2], help='rank')
    parser.add_argument('--use', type=int, default=2, help='use 2-3 or 2-3-4')

    parser.add_argument('--optim', type=str, default=opt.optim, help='Select the optimizer to use')
    parser.add_argument('--weight_decay', default=opt.weight_decay, help='')
    parser.add_argument('--momentum', type=float, default=opt.momentum, help='')

    parser.add_argument('--scheduler', type=str, default=opt.scheduler, help='Select the learning rate optimization to use')
    parser.add_argument('--max_lr', default=opt.max_lr, help='')
    parser.add_argument('--step', type=int, default=opt.step, help='')
    parser.add_argument('--gamma', default=opt.gamma, help='')
    parser.add_argument('--milestones', default=opt.milestones, help='')

    parser.add_argument('--lr', type=float, default=opt.lr, help='The learning rate used in training')
    parser.add_argument('--batch_size', type=int, default=opt.batch_size, help='The batch size used in training')
    parser.add_argument('--epoch', type=int, default=opt.epoch, help='The epoch used in training')
    parser.add_argument('--k', type=int, default=opt.k, help='topk')

    parser.add_argument('--plot_data', type=bool, default=opt.plot_data, help='Whether the dataset needs to be visualized')
    parser.add_argument('--plot_result', type=bool, default=opt.plot_result, help='Whether to visualize the classification results')
    parser.add_argument('--plot_confMatrix', type=bool, default=opt.plot_confMatrix, help='Whether you need to visualize the confusion matrix')
    parser.add_argument('--plot_kernel_feature', type=bool, default=opt.plot_kernel_feature, help='Whether you need to visualize the convolution kernel')
    parser.add_argument('--plot_feature_map', type=bool, default=opt.plot_feature_map, help='Whether to visualize the feature map')
    parser.add_argument('--send_message', type=bool, default=opt.send_message, help='Whether to send email after training')
    parser.add_argument('--save_weight', type=bool, default=opt.save_weight, help='Whether to save model parameters')
    parser.add_argument('--save_model', type=bool, default=opt.save_model, help='Whether to save the model')
    parser.add_argument('--tensorboard', type=bool, default=opt.tensorboard, help='Whether Tensorboard is required')
    parser.add_argument('--fine_tune', type=bool, default=opt.fine_tune, help='Whether the model needs fine-tuning')
    parser.add_argument('--frozen_params', type=bool, default=opt.frozen_params, help='Whether to freeze parameters')
    parser.add_argument('--init_params', type=bool, default=opt.init_params, help='')
    parser.add_argument('--device', type=str, default=opt.device, help='device id (i.e. 0 or 0,1 or cpu)')
    parser.add_argument('--train', default=opt.train, help='Whether to train the model')
    parser.add_argument('--load_weight', type=str, default=opt.load_weight, help='')
    parser.add_argument('--save_result', type=str, default=opt.save_result, help='Saving training results')
    parser.add_argument('--save_weight_path', type=str, default=opt.save_weight_path, help='Save model parameters')
    parser.add_argument('--save_tensorboard', type=str, default=opt.save_tensorboard, help='Save visualization results')
    parser.add_argument('--attention', default=opt.attention)

    parser.add_argument('--remark', type=str, default=opt.remark, help='any remark')

    args = parser.parse_args()

    return args


def main(args):

    #####################################################################################
    # todo [prepare log file]
    device = torch.device(args.device if torch.cuda.is_available() else "cpu")

    logger.create_dir(args.save_result)
    save_name = '{}_{}_{}_{}_log.txt'.format(time.strftime("%Y_%m_%d_%H:%M"), args.model_type, args.data_type,
                                             args.remark)
    log = logger.Logger(os.path.join(args.save_result, save_name))
    log.write(args.__repr__())

    #####################################################################################
    # todo [prepare dataset]
    print('loading data:', args.data_type)
    train_loader = train_data_load(args.data_type, args.batch_size)
    test_loader = test_data_load(args.data_type, args.batch_size)
    print(len(train_loader), len(test_loader))

    demenstrate.plot_data_loader_image(args, train_loader) if args.plot_data else None

    #####################################################################################
    # todo [prepare model]
    print('loading model:')
    if args.model_type == 'BSCP':
        model = BSCP_Resnet50(2048, 2048, args.rank, args.l1, args.l2, args.p, args.num_classes).to(args.device)
    elif args.model_type == 'BSCP_kernel':
        model = BSCP_kernelPooling(rank=args.RANK, l1=args.U, l2=args.V, p=args.P, num_classes=args.num_classes, use=args.use).to(args.device)
    else:
        exit(-1)

    # log.write('{}'.format(model))
    parameters = sum(x.numel() for x in model.parameters())
    log.write('The total parameters of the [{}] model are: {}'.format(args.model_type, parameters))

    #####################################################################################
    # todo [parameter setting]:load parameters/frozen parameters/init parameters
    op = None
    if args.load_weight != "":
        print("loading parameters:", args.load_weight)
        assert os.path.exists(args.load_weight), "weights file: '{}' not exist.".format(args.load_weight)
        weights_dict = torch.load(args.load_weight, map_location=args.device)

        # 删除不需要的权重
        del_keys = ['head.weight', 'head.bias'] if model.has_logits \
            else ['pre_logits.fc.weight', 'pre_logits.fc.bias', 'head.weight', 'head.bias']
        for k in del_keys:
            del weights_dict[k]

        print(model.load_state_dict(weights_dict, strict=False))

    if args.frozen_params:
        print("freezing parameters--")
        for name, para in model.named_parameters():
            if "fc" not in name:
                para.requires_grad = False
            else:
                log.write("\t training [{}]".format(name))
        op = [p for p in model.parameters() if p.requires_grad]
    else:
        for name, para in model.named_parameters():
            para.requires_grad = True
        log.write("\t training all")

    if args.init_params:
        None

    #####################################################################################
    # todo [prepare tensorboard/loss/optimizer/scheduler]
    if args.tensorboard:
        print("Use tensorboard---")
        writer = SummaryWriter(
            'runs/tensorboard/{}_{}_{}_{}'.format(time.strftime("%Y_%m_%d_%H:%M"), args.model_type, args.data_type,
                                                  args.remark))
        init_img = torch.zeros((1, 3, 224, 224), device=device)
        writer.add_graph(model, init_img)
    else:
        writer = None

    criterion = nn.CrossEntropyLoss()
    optimizer = optim_factory.create_optimizer(args, model, op)
    scheduler = scheduler_factory.create_scheduler(args, optimizer)
    log.write("{}".format(optimizer))
    log.write("{}:{}".format(args.scheduler, scheduler.state_dict()))

    #####################################################################################
    # todo [train/val]
    best_train_acc = 0
    best_acc = 0
    for epoch in range(args.epoch):
        start = time.time()
        # train
        train_loss, train_acc = train.train_one_epoch(model=model,
                                                      optimizer=optimizer,
                                                      data_loader=train_loader,
                                                      device=device,
                                                      epoch=epoch, loss_function=criterion)

        scheduler.step()

        # validate
        val_loss, val_acc, top5, confMatrix_x, confMatrix_y = train.evaluate(model=model,
                                                                             data_loader=test_loader,
                                                                             device=device,
                                                                             epoch=epoch, loss_function=criterion, k=args.k)

        end = time.time()

        total_prediction, total_recall, total_F1_score, result_table, geometric_mean, confMatrix = confusion.calculate_prediction_recall(
            confMatrix_x, confMatrix_y, args.num_classes)

        log.write(
            'Epoch:[{}/{}],Time:[{:.4f}s],lr:[{:.6f}/{}]--train_loss:[{:.4f}], train_acc:[{:.4f}], test_loss:[{:.4f}], '
            'test_acc(top1): [{:.4f}], top{}(test):[{:.4f}], Recall:[{:.4f}], Prediction:[{:.4f}], F1_Score:[{:.4f}], geometric_mean:[{:.4f}]'.format(
                (epoch + 1),
                args.epoch,
                (end - start),
                optimizer.state_dict()['param_groups'][0]['lr'],
                args.lr,
                train_loss,
                train_acc,
                val_loss,
                val_acc,
                args.k,
                top5,
                total_recall,
                total_prediction,
                total_F1_score,
                geometric_mean
            ))

        #####################################################################################
        # todo [save/visualization/email]
        if args.tensorboard:
            writer.add_scalar("train_loss", train_loss, epoch)
            writer.add_scalar("train_acc", train_acc, epoch)
            writer.add_scalar("test_loss", val_loss, epoch)
            writer.add_scalar("test_acc", val_acc, epoch)
            writer.add_scalar("lr", optimizer.param_groups[0]["lr"], epoch)

        if val_acc > best_acc:
            best_train_acc = train_acc
            best_acc = val_acc

            best_top5 = top5
            best_recall = total_recall
            best_prediction = total_prediction
            best_f1_score = total_F1_score
            best_geometric_mean = geometric_mean
            temp_table = result_table
            best_confMatrix = confMatrix

            if args.save_weight:
                sava_weight_name = "{}_{}_{}_{}.pth".format(time.strftime("%Y_%m_%d"), args.model_type, args.data_type,
                                                            args.remark)
                torch.save(model.state_dict(), os.path.join(args.save_weight_path, sava_weight_name))
            if args.save_model:
                sava_model_name = "{}_{}_{}{}.pkl".format(time.strftime("%Y_%m_%d"), args.model_type, args.data_type,
                                                           args.remark)
                torch.save(model, os.path.join(args.save_weight_path, sava_model_name))

    log.write(
        'The best acc is: Train[{}], Test[{}], Topk:[{}], Recall:[{}], Prediction:[{}], F1_score:[{}], stability:[{}]'.format(
            best_train_acc, best_acc, best_top5, best_recall, best_prediction, best_f1_score, best_geometric_mean))
    log.write('{}'.format(temp_table))

    if args.plot_confMatrix:
        confusion.plot_confusion_matrix(best_confMatrix, list(range(args.num_classes)))

    if args.send_message:
        message = '{}:The {} experiment ended! \n\tThe experimental parameters are {}.' \
                  '\n\tAnd the best acc is {}%.'.format(time.strftime("%Y_%m_%d_%H:%M"), args.model_type, args.__repr__(),
                                                        best_acc)
        send_email.sendTo(message, args.model_type, best_acc)


if __name__ == '__main__':
    args = parse_args()

    # todo [choose train]
    if args.train:
        main(args)

    # todo [choose fine tune]
    elif args.fine_tune:
        print("fine tuning")
        fine_tuning(args, weight='weights/train_weight/2022_09_14_BCNN_CUB_step1.pth')           # todo

    # todo [visualization:result/kernel feature/feature map]
    else:
        device = torch.device(args.device if torch.cuda.is_available() else "cpu")
        print("loading model with parameters--")
        # todo
        # model = torch.load('weights/train_weight/2022_09_13_ViT_CUB.pkl')       # 加载模型
        model = BSCP_Resnet50(2048, 2048, 2, 512, 512, 512, 200).to(args.device)

        # model = atten_BSCP_Resnet50(2048, 2048, 2, 512, 512, 2048, 200).to(device)
        # model.load_state_dict(torch.load('weights/train_weight/2022_09_16_BCNN_CUB_.pth'))
        # model.load_state_dict(torch.load('weights/train_weight/2022_09_16_sp_CUB_sp.pth'))

        # test_loader = sample_test_data_load(args.data_type, args.batch_size)
        test_loader = sample_test_data_load(args.data_type, 6)
        # test_loader = test_data_load(args.data_type, args.batch_size)

        if args.plot_result:        # 给定模型和数据即可
            print("\t show result")
            demenstrate.plot_classification_result(args, test_loader, model)

        if args.plot_kernel_feature:
            print("\t show kernel feature")
            plot_feature.plot_kernel_feature(model)

        if args.plot_feature_map:       # 需要提前指定查看feature map的层数
            print("\t show feature map")
            map = {}

            def get_map(name):
                def hook(model, input, output):
                    map[name] = output.detach()
                return hook
            model.eval()

            # todo [which layer feature map to select]
            model.features[7][2].register_forward_hook(get_map('conv2d'))
            # model.bscp.attend.register_forward_hook(get_map('conv2d'))

            plot_feature.plot_feature_map(args, model, test_loader, map)

        if args.attention:
            print("\t show attend map")
            map = {}

            def get_map(name):
                def hook(model, input, output):
                    map[name] = output.detach()

                return hook

            model.eval()

            # model.bscp.attend.register_forward_hook(get_map('conv2d'))
            # model.bscp.s.register_forward_hook(get_map('conv2d'))
            model.spatial_attend_2.attend.register_forward_hook(get_map('conv2d'))

            plot_feature.plt_attention_map(args, model, test_loader, map, type='qkv')



