#!/usr/bin/python3.6

import sys
import torch
from tqdm import tqdm
"""train and eval"""


def train_one_epoch(model, optimizer, data_loader, device, epoch, loss_function=torch.nn.CrossEntropyLoss()):
    model.train()
    accu_loss = torch.zeros(1).to(device)  # 累计损失
    accu_num = torch.zeros(1).to(device)   # 累计预测正确的样本数
    optimizer.zero_grad()

    sample_num = 0
    data_loader = tqdm(data_loader, file=sys.stdout)
    for step, data in enumerate(data_loader):
        images, labels = data
        sample_num += images.shape[0]

        pred = model(images.to(device))
        pred_classes = torch.max(pred, dim=1)[1]
        accu_num += torch.eq(pred_classes, labels.to(device)).sum()

        loss = loss_function(pred, labels.to(device))
        loss.backward()
        accu_loss += loss.detach()

        data_loader.desc = "[train epoch:{}] train_loss: {:.3f}, train_acc: {:.3f}".format(epoch + 1,
                                                                               accu_loss.item() / (step + 1),
                                                                               accu_num.item() / sample_num)

        if not torch.isfinite(loss):
            print(' WARNING: non-finite loss, ending training', loss)
            sys.exit(1)

        optimizer.step()
        optimizer.zero_grad()

    return accu_loss.item() / (step + 1), accu_num.item() / sample_num


@torch.no_grad()
def evaluate(model, data_loader, device, epoch, loss_function=torch.nn.CrossEntropyLoss(), k=5):

    model.eval()

    accu_num = torch.zeros(1).to(device)   # 累计预测正确的样本数
    accu_loss = torch.zeros(1).to(device)  # 累计损失

    confMatrix_x = torch.empty(1)
    confMatrix_y = torch.empty(1)
    topk = 0

    sample_num = 0
    data_loader = tqdm(data_loader, file=sys.stdout)
    for step, data in enumerate(data_loader):
        images, labels = data
        sample_num += images.shape[0]

        pred = model(images.to(device))
        pred_classes = torch.max(pred, dim=1)[1]
        accu_num += torch.eq(pred_classes, labels.to(device)).sum()

        loss = loss_function(pred, labels.to(device))
        accu_loss += loss

        _, max_k = torch.topk(pred, k, dim=-1)
        test_labels = labels.view(-1, 1)
        topk += (test_labels.cpu() == max_k.cpu()).sum().item()

        confMatrix_x = torch.cat((confMatrix_x, labels.cpu()), dim=0)
        confMatrix_y = torch.cat((confMatrix_y, pred_classes.cpu()), dim=0)

        data_loader.desc = "[valid epoch:{}] valid_loss: {:.3f}, valid_acc: {:.3f}, top{}: {:.3f}".format(epoch + 1,
                                                                               accu_loss.item() / (step + 1),
                                                                               accu_num.item() / sample_num,
                                                                               k,
                                                                               topk / sample_num)

    return accu_loss.item() / (step + 1), accu_num.item() / sample_num, topk / sample_num, confMatrix_x, confMatrix_y
