
import torch
from torch import nn
from utils import confusion, demenstrate, logger, optim_factory, plot_feature, scheduler_factory, send_email, tools, train
import time
import os
from datasets import train_data_load, test_data_load
from torch.utils.tensorboard import SummaryWriter
from model.LightW_BSCP import LightBSCPResnet50


def fine_tuning(args, weight=''):
    device = torch.device(args.device if torch.cuda.is_available() else "cpu")

    logger.create_dir(args.save_result)
    save_name = '{}_{}_{}_finetune_log.txt'.format(time.strftime("%Y_%m_%d_%H:%M"), args.model_type, args.data_type)
    log = logger.Logger(os.path.join(args.save_result, save_name))
    log.write(args.__repr__())

    print('loading data:', args.data_type)
    train_loader = train_data_load(args.data_type, args.batch_size)
    test_loader = test_data_load(args.data_type, args.batch_size)
    print(len(train_loader), len(test_loader))

    demenstrate.plot_data_loader_image(args, train_loader) if args.plot_data else None

    ##########################
    # todo
    print('loading model:')
    model = LightBSCPResnet50(in_channel=2048, out_channel=args.out_channel, out_dim=args.out_dim, rank=args.rank,
                              num_classes=args.num_classes).to(device)

    # log.write('{}'.format(model))
    parameters = sum(x.numel() for x in model.parameters())
    log.write('The total parameters of the [{}] model are: {}'.format(args.model_type, parameters))

    ###########################

    print("loading parameters:", weight)
    assert os.path.exists(weight), "weights file: '{}' not exist.".format(weight)
    weights_dict = torch.load(weight, map_location=args.device)
    print(model.load_state_dict(weights_dict, strict=False))

    ###########################

    if args.tensorboard:
        print("Use tensorboard---")
        writer = SummaryWriter('runs/tensorboard/{}_{}_{}_finetune'.format(time.strftime("%Y_%m_%d_%H:%M"), args.model_type, args.data_type))
        init_img = torch.zeros((1, 3, 224, 224), device=device)
        writer.add_graph(model, init_img)
    else:
        writer = None

    criterion = nn.CrossEntropyLoss()
    optimizer = optim_factory.create_optimizer(args, model)
    scheduler = scheduler_factory.create_scheduler(args, optimizer)
    log.write("{}".format(optimizer))
    log.write("{}:{}".format(args.scheduler, scheduler.state_dict()))

    #############################

    best_train_acc = 0
    best_acc = 0
    for epoch in range(args.epoch):
        start = time.time()
        # train
        train_loss, train_acc = train.train_one_epoch(model=model,
                                                      optimizer=optimizer,
                                                      data_loader=train_loader,
                                                      device=device,
                                                      epoch=epoch, loss_function=criterion)

        scheduler.step()

        # validate
        val_loss, val_acc, top5, confMatrix_x, confMatrix_y = train.evaluate(model=model,
                                                                             data_loader=test_loader,
                                                                             device=device,
                                                                             epoch=epoch, loss_function=criterion,
                                                                             k=args.k)

        end = time.time()

        total_prediction, total_recall, total_F1_score, result_table, geometric_mean, confMatrix = confusion.calculate_prediction_recall(
            confMatrix_x, confMatrix_y, args.num_classes)

        log.write(
            'Epoch:[{}/{}],Time:[{:.4f}s],lr:[{:.6f}/{}]--train_loss:[{:.4f}], train_acc:[{:.4f}], test_loss:[{:.4f}], '
            'test_acc(top1): [{:.4f}], top{}(test):[{:.4f}], Recall:[{:.4f}], Prediction:[{:.4f}], F1_Score:[{:.4f}], geometric_mean:[{:.4f}]'.format(
                (epoch + 1),
                args.epoch,
                (end - start),
                optimizer.state_dict()['param_groups'][0]['lr'],
                args.lr,
                train_loss,
                train_acc,
                val_loss,
                val_acc,
                args.k,
                top5,
                total_recall,
                total_prediction,
                total_F1_score,
                geometric_mean
            ))

        ############################

        if args.tensorboard:
            writer.add_scalar("train_loss", train_loss, epoch)
            writer.add_scalar("train_acc", train_acc, epoch)
            writer.add_scalar("test_loss", val_loss, epoch)
            writer.add_scalar("test_acc", val_acc, epoch)
            writer.add_scalar("lr", optimizer.param_groups[0]["lr"], epoch)

        if val_acc > best_acc:
            best_train_acc = train_acc
            best_acc = val_acc

            best_top5 = top5
            best_recall = total_recall
            best_prediction = total_prediction
            best_f1_score = total_F1_score
            best_geometric_mean = geometric_mean
            temp_table = result_table
            best_confMatrix = confMatrix

            if args.save_weight:
                sava_weight_name = "{}_{}_{}_finetune.pth".format(time.strftime("%Y_%m_%d"), args.model_type, args.data_type)
                torch.save(model.state_dict(), os.path.join(args.save_weight_path, sava_weight_name))
            if args.save_model:
                sava_model_name = "{}_{}_{}_finetune.pkl".format(time.strftime("%Y_%m_%d"), args.model_type, args.data_type)
                torch.save(model, os.path.join(args.save_weight_path, sava_model_name))

    log.write(
        'The best acc is: Train[{}], Test[{}], Topk:[{}], Recall:[{}], Prediction:[{}], F1_score:[{}], stability:[{}]'.format(
            best_train_acc, best_acc, best_top5, best_recall, best_prediction, best_f1_score, best_geometric_mean))
    log.write('{}'.format(temp_table))

    if args.plot_confMatrix:
        confusion.plot_confusion_matrix(best_confMatrix, list(range(args.num_classes)))

    if args.send_message:
        message = '{}:The {} experiment ended! \n\tThe experimental parameters are {}.' \
                  '\n\tAnd the best acc is {}%.'.format(time.strftime("%Y_%m_%d_%H:%M"), args.model_type,
                                                        args.__repr__(),
                                                        best_acc)
        send_email.sendTo(message, args.model_type, best_acc)
